"""
Reconsitutes an ontology from SPARQL queries over a remote SPARQL server

 * the first time an ontology is referenced, basic axioms will be fetched via SPARQL
 * these will be cached in `/tmp/.cache/`
 * the second time the same ontology is referenced, the disk cache will be used
 * if an ontology is referenced a second time in the same in-memory session, the disk cache is bypassed and in-memory (lru) cache is used

Note: you should not need to use this directly. An `OntologyFactory` object will automatically use this if a sparql handle is passed.

"""

from SPARQLWrapper import SPARQLWrapper, JSON
from prefixcommons.curie_util import contract_uri, expand_uri
from ontobio.vocabulary.relations import map_legacy_pred
from functools import lru_cache
import networkx
import logging

from enum import Enum

SEPARATOR = "@|@"


# CACHE STRATEGY:
# by default, the cache is NOT persistent. Only single threaded clients should
# call a cached method with writecache=True.
# Note we are layering the in-memory cache over the persistent cache

cache = lru_cache(maxsize=None)


SUBCLASS_OF = 'subClassOf'
SUBPROPERTY_OF = 'subPropertyOf'

logger = logging.getLogger(__name__)


class EOntology(Enum):
    GO = "http://rdf.geneontology.org/sparql"
    HEGROUP = "http://sparql.hegroup.org/sparql"
    UNIPROT = "https://sparql.uniprot.org/sparql/"

class EDocState(Enum):
    PRODUCTION = "production"
    DEVELOPMENT = "development"
    REVIEW = "review"
    DELETE = "delete"

# TODO
# for now we assume ontobee
ontol_sources = {
    'go': "http://rdf.geneontology.org/sparql",
    '': "http://sparql.hegroup.org/sparql"
    }

OIO_SYNS = dict(
    related='oboInOwl:hasRelatedSynonym',
    narrow = 'oboInOwl:hasNarrowSynonym',
    broad = 'oboInOwl:hasBroadSynonym',
    exact = 'oboInOwl:hasExactSynonym')
    

def get_digraph(ont, relations=None, writecache=False):
    """
    Creates a basic graph object corresponding to a remote ontology
    """
    digraph = networkx.MultiDiGraph()
    logger.info("Getting edges (may be cached)")
    for (s,p,o) in get_edges(ont):
        p = map_legacy_pred(p)
        if relations is None or p in relations:
            digraph.add_edge(o,s,pred=p)
    logger.info("Getting labels (may be cached)")
    for (n,label) in fetchall_labels(ont):
        digraph.add_node(n, **{'label':label})
    return digraph

def get_xref_graph(ont):
    """
    Creates a basic graph object corresponding to a remote ontology
    """
    g = networkx.MultiGraph()
    for (c,x) in fetchall_xrefs(ont):
        g.add_edge(c,x,source=c)
    return g


def get_edges(ont):
    """
    Fetches all basic edges from a remote ontology
    """
    logger.info("QUERYING:"+ont)
    edges = [(c,SUBCLASS_OF, d) for (c,d) in fetchall_isa(ont)]
    edges += fetchall_svf(ont)
    edges += [(c,SUBPROPERTY_OF, d) for (c,d) in fetchall_subPropertyOf(ont)]
    if len(edges) == 0:
        logger.warning("No edges for {}".format(ont))
    return edges

def search(ont, searchterm):
    """
    Search for things using labels
    """
    namedGraph = get_named_graph(ont)
    query = """
    SELECT ?c ?l WHERE {{
    GRAPH <{g}>  {{
    ?c rdfs:label ?l
    FILTER regex(?l,'{s}','i')
    }}
    }}
    """.format(s=searchterm, g=namedGraph)
    bindings = run_sparql(query)
    return [(r['c']['value'],r['l']['value']) for r in bindings]

def get_terms_in_subset(ont, subset):
    """
    Find all nodes in a subset.

    We assume the oboInOwl encoding of subsets, and subset IDs are IRIs
    """
    namedGraph = get_named_graph(ont)

    # note subsets have an unusual encoding
    query = """
    prefix oboInOwl: <http://www.geneontology.org/formats/oboInOwl#>
    SELECT ?c ? WHERE {{
    GRAPH <{g}>  {{
    ?c oboInOwl:inSubset ?s ;
       rdfs:label ?l
    FILTER regex(?s,'#{s}$','i')
    }}
    }}
    """.format(s=subset, g=namedGraph)
    bindings = run_sparql(query)
    return [(r['c']['value'],r['l']['value']) for r in bindings]


def run_sparql(q):
    # TODO: select endpoint based on ontology
    #sparql = SPARQLWrapper("http://rdf.geneontology.org/sparql")
    logger.info("Connecting to sparql endpoint...")
    sparql = SPARQLWrapper("http://sparql.hegroup.org/sparql")
    logger.info("Made wrapper: {}".format(sparql))
    # TODO: iterate over large sets?
    full_q = q + ' LIMIT 250000'
    sparql.setQuery(q)
    sparql.setReturnFormat(JSON)
    logger.info("Query: {}".format(q))
    results = sparql.query().convert()
    bindings = results['results']['bindings']
    logger.info("Rows: {}".format(len(bindings)))
    for r in bindings:
        curiefy(r)
    return bindings


def run_sparql_on(q, ontology):
    """
    Run a SPARQL query (q) on a given Ontology (Enum EOntology)
    """
    logger.info("Connecting to " + ontology.value + " SPARQL endpoint...")
    sparql = SPARQLWrapper(ontology.value)
    logger.info("Made wrapper: {}".format(sparql))
    sparql.setQuery(q)
    sparql.setReturnFormat(JSON)
    logger.info("Query: {}".format(q))
    results = sparql.query().convert()
    bindings = results['results']['bindings']
    logger.info("Rows: {}".format(len(bindings)))
    for r in bindings:
        curiefy(r)
    return bindings


def curiefy(r):
    for (k,v) in r.items():
        if v['type'] == 'uri':
            curies = contract_uri(v['value'])
            if len(curies)>0:
                r[k]['value'] = curies[0]
                
def get_named_graph(ont):
    """
    Ontobee uses NGs such as http://purl.obolibrary.org/obo/merged/CL
    """

    if ont.startswith('http://'):
        return ont
    namedGraph = 'http://purl.obolibrary.org/obo/merged/' + ont.upper()
    return namedGraph

def fetchall_isa(ont):
    namedGraph = get_named_graph(ont)
    queryBody = querybody_isa()
    query = """
    SELECT * WHERE {{
    GRAPH <{g}>  {q}
    }}
    """.format(q=queryBody, g=namedGraph)
    bindings = run_sparql(query)
    return [(r['c']['value'],r['d']['value']) for r in bindings]

def fetchall_subPropertyOf(ont):
    namedGraph = get_named_graph(ont)
    queryBody = querybody_subPropertyOf()
    query = """
    SELECT * WHERE {{
    GRAPH <{g}>  {q}
    }}
    """.format(q=queryBody, g=namedGraph)
    bindings = run_sparql(query)
    return [(r['c']['value'],r['d']['value']) for r in bindings]

def fetchall_svf(ont):
    namedGraph = get_named_graph(ont)
    queryBody = querybody_svf()
    query = """
    SELECT * WHERE {{
    GRAPH <{g}>  {q}
    }}
    """.format(q=queryBody, g=namedGraph)
    bindings = run_sparql(query)
    return [(r['c']['value'], r['p']['value'], r['d']['value']) for r in bindings]

def fetchall_labels(ont):
    """
    fetch all rdfs:label assertions for an ontology
    """
    logger.info("fetching rdfs:labels for: "+ont)
    namedGraph = get_named_graph(ont)
    queryBody = querybody_label()
    query = """
    SELECT * WHERE {{
    GRAPH <{g}>  {q}
    }}
    """.format(q=queryBody, g=namedGraph)
    bindings = run_sparql(query)
    rows = [(r['c']['value'], r['l']['value']) for r in bindings]
    return rows

def fetchall_syns(ont):
    """
    fetch all synonyms for an ontology
    """
    logger.info("fetching syns for: "+ont)
    namedGraph = get_named_graph(ont)
    queryBody = querybody_syns()
    query = """
    prefix oboInOwl: <http://www.geneontology.org/formats/oboInOwl#>
    SELECT * WHERE {{
    GRAPH <{g}>  {q}
    }}
    """.format(q=queryBody, g=namedGraph)
    bindings = run_sparql(query)
    rows = [(r['c']['value'], r['r']['value'], r['l']['value']) for r in bindings]
    return rows

def fetchall_textdefs(ont):
    """
    fetch all text defs for an ontology
    """
    logger.info("fetching text defs for: "+ont)
    namedGraph = get_named_graph(ont)
    query = """
    prefix IAO: <http://purl.obolibrary.org/obo/IAO_>
    SELECT * WHERE {{
    GRAPH <{g}>  {{
      ?c IAO:0000115 ?d
    }}
    FILTER (!isBlank(?c))
    }}
    """.format(g=namedGraph)
    bindings = run_sparql(query)
    rows = [(r['c']['value'], r['d']['value']) for r in bindings]
    return rows

def fetchall_xrefs(ont):
    """
    fetch all xrefs for an ontology
    """
    logger.info("fetching xrefs for: "+ont)
    namedGraph = get_named_graph(ont)
    query = """
    prefix oboInOwl: <http://www.geneontology.org/formats/oboInOwl#>
    SELECT * WHERE {{
    GRAPH <{g}>  {{ ?c oboInOwl:hasDbXref ?x }}
    FILTER (!isBlank(?c))
    }}
    """.format(g=namedGraph)
    bindings = run_sparql(query)
    rows = [(r['c']['value'], r['x']['value']) for r in bindings]
    return rows

def fetchall_obs(ont):
    """
    fetch all obsoletes for an ontology
    """
    logger.info("fetching obsoletes for: "+ont)
    namedGraph = get_named_graph(ont)
    query = """
    SELECT ?c WHERE {{
    GRAPH <{g}>  {{ ?c owl:deprecated "true"^^xsd:boolean }}
    FILTER (!isBlank(?c))
    }}
    """.format(g=namedGraph)
    bindings = run_sparql(query)
    rows = [r['c']['value'] for r in bindings]
    return rows

def querybody_isa():
    return """
    { ?c rdfs:subClassOf ?d }
    FILTER (!isBlank(?c))
    FILTER (!isBlank(?d))
    """

def querybody_subPropertyOf():
    return """
    { ?c rdfs:subPropertyOf ?d }
    FILTER (!isBlank(?c))
    FILTER (!isBlank(?d))
    """

def querybody_svf():
    return """
    { ?c rdfs:subClassOf [owl:onProperty ?p ; owl:someValuesFrom ?d ] }
    FILTER (!isBlank(?c))
    FILTER (!isBlank(?p))
    FILTER (!isBlank(?d))
    """

def querybody_label():
    return """
    { ?c rdfs:label ?l
    FILTER (!isBlank(?c))}
    """

def querybody_syns():
    return """
    { ?c ?r ?l }
    FILTER (
    ?r = oboInOwl:hasRelatedSynonym OR
    ?r = oboInOwl:hasNarrowSynonym OR
    ?r = oboInOwl:hasBroadSynonym OR
    ?r = oboInOwl:hasExactSynonym
    )
    FILTER (!isBlank(?c))
    """


def anyont_fetch_label(id):
    """
    fetch all rdfs:label assertions for a URI
    """
    iri = expand_uri(id, strict=False)
    query = """
    SELECT ?label WHERE {{
    <{iri}> rdfs:label ?label
    }}
    """.format(iri=iri)
    bindings = run_sparql(query)
    rows = [r['label']['value'] for r in bindings]
    return rows[0]

def batch_fetch_labels(ids):
    """
    fetch all rdfs:label assertions for a set of CURIEs
    """
    m = {}
    for id in ids:
        label = anyont_fetch_label(id)
        if label is not None:
            m[id] = label
    return m






def transform(data, keysToSplit=[]):
    """
    Transform a SPARQL json result by:
    1) outputing only { key : value }, removing datatype
    2) for some keys, transform them into array based on SEPARATOR
    """
    transformed = { }
    for key in data:
        if key in keysToSplit:
            transformed[key] = data[key]['value'].split(SEPARATOR)
        else:
            transformed[key] = data[key]['value']
    return transformed

def transformArray(data, keysToSplit=[]):
    """
    Transform a SPARQL json array based on the rules of transform
    """
    transformed = [ ]
    for item in data:
        transformed.append(transform(item, keysToSplit))
    return transformed

    
## -- PERSISTENCE --





"""
Convenience wrapper for golr_query.py

In general you should use the `GolrAssociationQuery` object directly. This module provide convenience non-OO functions that typically compose the creation of a GolrAssociationQuery object followed by the `exec` method.

"""
import logging

import pysolr
import json
import logging
import time
from ontobio.golr.golr_query import *
from ontobio.util.user_agent import get_user_agent

import math ### required for IC calculation

MAX_ROWS = 100000

logger = logging.getLogger(__name__)

def get_association(id, **kwargs):
    """
    Fetch an association object by ID
    """
    results = search_associations(id=id, **kwargs)
    assoc = results['associations'][0] if len(results['associations']) > 0 else {}
    return assoc

def search_associations(**kwargs):

    """
    Fetch a set of association objects based on a query.
    :rtype:
    """
    logger.info("CREATING_GOLR_QUERY {}".format(kwargs))
    q = GolrAssociationQuery(**kwargs)
    return q.exec()

def get_objects_for_subject(subject=None,
                            object_category=None,
                            relation=None,
                            **kwargs):
    """
    Convenience method: Given a subject (e.g. gene, disease, variant), return all associated objects (phenotypes, functions, interacting genes, etc)
    """
    searchresult = search_associations(subject=subject,
                                       fetch_objects=True,
                                       rows=0,
                                       object_category=object_category,
                                       relation=relation,
                                       **kwargs
    )
    objs = searchresult['objects']
    return objs

def get_subjects_for_object(object=None,
                            subject_category=None,
                            subject_taxon=None,
                            relation=None,
                            **kwargs):
    """
    Convenience method: Given a object (e.g. ontology term like phenotype or GO; interacting gene; disease; pathway etc), return all associated subjects (genes, variants, pubs, etc)
    """
    searchresult = search_associations(object=object,
                                       fetch_subjects=True,
                                       rows=0,
                                       subject_category=subject_category,
                                       subject_taxon=subject_taxon,
                                       relation=relation,
                                       **kwargs
    )
    subjs = searchresult['subjects']
    return subjs

def search_associations_compact(**kwargs):
    """
    Convenience method: as for search associations, use compact
    """
    searchresult = search_associations(use_compact_associations=True,
                                       facet_fields=[],
                                       **kwargs
    )
    return searchresult['compact_associations']

def map2slim(subjects, slim, **kwargs):
    """
    Maps a set of subjects (e.g. genes) to a set of slims

    Result is a list of unique subject-class pairs, with
    a list of source assocations
    """
    logger.info("SLIM SUBJECTS:{} SLIM:{} CAT:{}".format(subjects,slim,kwargs.get('category')))
    searchresult = search_associations(subjects=subjects,
                                       slim=slim,
                                       facet_fields=[],
                                       **kwargs
    )
    pmap = {}
    for a in searchresult['associations']:
        subj = a['subject']['id']
        slimmed_terms = a['slim']
        #logger.info('SLIM: {} {}'.format(subj,slimmed_terms))
        for t in slimmed_terms:
            k = (subj,t)
            if k not in pmap:
                pmap[k] = []
            pmap[k].append(a)
    results = [ {'subject': subj, 'slim':t, 'assocs': assocs} for ((subj,t),assocs) in pmap.items()]
    return results

def top_species(**kwargs):
    results = search_associations(facet_fields = [M.SUBJECT_TAXON],
                                  facet=True,
                                  facet_limit=-1,
                                  rows=0,
                                  **kwargs)
    fcs = results['facet_counts']
    logger.info("FCs={}".format(fcs))
    ## TODO:
    if 'taxon' in fcs:
        d = fcs['taxon']
    else:
        d = fcs['subject_taxon']
    return sorted(d.items(), key=lambda t: -t[1])
    


def bulk_fetch(subject_category, object_category, taxon, rows=MAX_ROWS, **kwargs):
    """
    Fetch associations for a species and pair of categories in bulk.

    Arguments:

     - subject_category: String (not None)
     - object_category: String (not None)
     - taxon: String
     - rows: int

    Additionally, any argument for search_associations can be passed
    """
    assert subject_category is not None
    assert object_category is not None
    time.sleep(1)
    logger.info("Bulk query: {} {} {}".format(subject_category, object_category, taxon))
    assocs = search_associations_compact(subject_category=subject_category,
                                         object_category=object_category,
                                         subject_taxon=taxon,
                                         rows=rows,
                                         iterate=True,
                                         **kwargs)
    logger.info("Rows retrieved: {}".format(len(assocs)))
    if len(assocs) == 0:
        logger.error("No associations returned for query: {} {} {}".format(subject_category, object_category, taxon))
    return assocs

def pivot_query(facet=None, facet_pivot_fields=None, **kwargs):
    """
    Pivot query
    """
    if facet_pivot_fields is None:
        facet_pivot_fields = []
    results = search_associations(rows=0,
                                  facet_fields=[facet],
                                  #facet_pivot_fields=facet_pivot_fields + [facet],
                                  facet_pivot_fields=facet_pivot_fields,
                                  **kwargs)
    return results

def pivot_query_as_matrix(facet=None, facet_pivot_fields=None, **kwargs):
    """
    Pivot query
    """
    if facet_pivot_fields is None:
        facet_pivot_fields = []
    logger.info("Additional args: {}".format(kwargs))
    fp = search_associations(rows=0,
                             facet_fields=[facet],
                             facet_pivot_fields=facet_pivot_fields,
                             **kwargs)['facet_pivot']

    # we assume only one
    results = list(fp.items())[0][1]
    tups = []
    xtype=None
    ytype=None
    xlabels=set()
    ylabels=set()

    
    for r in results:
        logger.info("R={}".format(r))
        xtype=r['field']
        rv = r['value']
        xlabels.add(rv)
        for piv in r['pivot']:
            ytype=piv['field']
            pv = piv['value']
            ylabels.add(pv)
            tups.append( (rv,pv,piv['count']) )

    z = [ [0] * len(xlabels) for i1 in range(len(ylabels)) ]

    xlabels=list(xlabels)
    ylabels=list(ylabels)
    xmap = dict([x[::-1] for x in enumerate(xlabels)])
    ymap = dict([x[::-1] for x in enumerate(ylabels)])
    for t in tups:
        z[ymap[t[1]]][xmap[t[0]]] = t[2]
    m = {'xtype':xtype,
         'ytype':ytype,
         'xaxis':xlabels,
         'yaxis':ylabels,
         'z':z}
    return m
         


# TODO: unify this with the monarch-specific instance
# note that longer term the goal is to unify the go and mon
# golr schemas more. For now the simplest path is
# to introduce this extra method, and 'mimic' the monarch one,
# at the risk of some duplication of code and inelegance

def search_associations_go(
        subject_category=None,
        object_category=None,
        relation=None,
        subject=None,
        **kwargs):
    """
    Perform association search using Monarch golr
    """
    go_golr_url = "http://golr.geneontology.org/solr/"
    go_solr = pysolr.Solr(go_golr_url, timeout=5)
    go_solr.get_session().headers['User-Agent'] = get_user_agent(caller_name=__name__)
    return search_associations(subject_category,
                               object_category,
                               relation,
                               subject,
                               solr=go_solr,
                               field_mapping=goassoc_fieldmap(),
                               **kwargs)

def select_distinct(distinct_field=None, **kwargs):
    """
    select distinct values for a given field for a given a query
    """
    results = search_associations(rows=0,
                                  select_fields=[],
                                  facet_field_limits = {
                                      distinct_field : -1
                                  },
                                  facet_fields=[distinct_field],
                                  **kwargs
    )
    # TODO: map field
    return list(results['facet_counts'][distinct_field].keys())


def select_distinct_subjects(**kwargs):
    """
    select distinct subject IDs given a query
    """
    return select_distinct(M.SUBJECT, **kwargs)

def calculate_information_content(**kwargs):
    """

    Arguments are as for search_associations, in particular:

     - subject_category
     - object_category
     - subject_taxon

    """
    # TODO - constraint using category and species
    results = search_associations(rows=0,
                                  select_fields=[],
                                  facet_field_limits = {
                                      M.OBJECT : -1
                                  },
                                  facet_fields=[M.OBJECT],
                                  **kwargs
    )
    pop_size = None
    icmap = {}

    # find max
    for (f,fc) in results['facet_counts'][M.OBJECT].items():
        if pop_size is None or pop_size < fc:
            pop_size = fc

    # IC = -Log2(freq)
    for (f,fc) in results['facet_counts'][M.OBJECT].items():
        freq = fc/pop_size
        icmap[f] = -math.log(freq, 2)
    return icmap

from ontobio.vocabulary.relations import HomologyTypes

def get_homologs(gene, relation=HomologyTypes.Ortholog.value, **kwargs):
    return search_associations(subject_category='gene',
                        object_category='gene',
                        relation=relation,
                        subject=gene,
                        **kwargs)




"""
Factory class for generating ontology objects based on a variety of handle types.

See :ref:`inputs` on readthedocs for more details
"""

import ontobio.obograph_util as obograph_util
from ontobio.ontol import Ontology
from ontobio.sparql.sparql_ontology import EagerRemoteSparqlOntology
import os
import subprocess
import hashlib
import logging

logger = logging.getLogger(__name__)

# TODO
default_ontology_handle = 'cache/ontologies/pato.json'
#if not os.path.isfile(ontology_handle):
#    ontology_handle = None

global default_ontology
default_ontology = None


class OntologyFactory():
    """Implements a factory for generating :class:`Ontology` objects.

    You should use a factory object rather than initializing
    `Ontology` directly. See :ref:`inputs` for more details.
    """

    # class variable - reuse the same object throughout
    test = 0

    def __init__(self, handle=None):
        """
        initializes based on an ontology name

        Arguments
        ---------
        handle : str
            see `create`
        """
        self.handle = handle

    def create(self, handle=None, handle_type=None, **args):
        """
        Creates an ontology based on a handle

        Handle is one of the following

        - `FILENAME.json` : creates an ontology from an obographs json file
        - `obo:ONTID`     : E.g. obo:pato - creates an ontology from obolibrary PURL (requires owltools)
        - `ONTID`         : E.g. 'pato' - creates an ontology from a remote SPARQL query

        Arguments
        ---------
        handle : str
            specifies how to retrieve the ontology info

        """
        if handle is None:
            self.test = self.test+1
            logger.info("T: "+str(self.test))
            global default_ontology
            if default_ontology is None:
                logger.info("Creating new instance of default ontology")
                default_ontology = create_ontology(default_ontology_handle, **args)
            logger.info("Using default_ontology")
            return default_ontology
        return create_ontology(handle, **args)

def create_ontology(handle=None, **args):
    ont = None
    logger.info("Determining strategy to load '{}' into memory...".format(handle))

    if handle.find("+") > -1:
        handles = handle.split("+")
        onts = [create_ontology(ont) for ont in handles]
        ont = onts.pop()
        ont.merge(onts)
        return ont

    # TODO: consider replacing with plugin architecture
    if handle.find(".") > 0 and os.path.isfile(handle):
        logger.info("Fetching obograph-json file from filesystem")
        ont = translate_file_to_ontology(handle, **args)
    elif handle.startswith("obo:"):
        logger.info("Fetching from OBO PURL")
        if handle.find(".") == -1:
            if handle == 'chebi' or handle == 'ncbitaxon' or handle == 'pr':
                handle += '.obo'
                logger.info("using obo for large ontology: {}".format(handle))
            else:
                handle += '.owl'
        fn = '/tmp/'+handle
        if not os.path.isfile(fn):
            url = handle.replace("obo:","http://purl.obolibrary.org/obo/")
            cmd = ['owltools',url,'-o','-f','json',fn]
            cp = subprocess.run(cmd, check=True)
            logger.info(cp)
        else:
            logger.info("using cached file: "+fn)
        g = obograph_util.convert_json_file(fn)
        ont = Ontology(handle=handle, payload=g)
    elif handle.startswith("wdq:"):
        from ontobio.sparql.wikidata_ontology import EagerWikidataOntology
        logger.info("Fetching from Wikidata")
        ont = EagerWikidataOntology(handle=handle)
    elif handle.startswith("skos:"):
        fn = handle.replace('skos:','')
        from ontobio.sparql.skos import Skos
        logger.info("Fetching from Skos file")
        skos = Skos()
        ont = skos.process_file(fn)
    elif handle.startswith("scigraph:"):
        from ontobio.neo.scigraph_ontology import RemoteScigraphOntology
        logger.info("Fetching from SciGraph")
        ont = RemoteScigraphOntology(handle=handle)
    elif handle.startswith("http:"):
        logger.info("Fetching from Web PURL: "+handle)
        encoded = hashlib.sha256(handle.encode()).hexdigest()
        #encoded = binascii.hexlify(bytes(handle, 'utf-8'))
        #base64.b64encode(bytes(handle, 'utf-8'))
        logger.info(" encoded: "+str(encoded))
        fn = '/tmp/'+encoded
        if not os.path.isfile(fn):
            cmd = ['owltools',handle,'-o','-f','json',fn]
            cp = subprocess.run(cmd, check=True)
            logger.info(cp)
        else:
            logger.info("using cached file: "+fn)
        g = obograph_util.convert_json_file(fn)
        ont = Ontology(handle=handle, payload=g)
    else:
        logger.info("Fetching from SPARQL")
        ont = EagerRemoteSparqlOntology(handle=handle)
        #g = get_digraph(handle, None, True)
    return ont

def create_ontology_from_obograph(og):
    ont = None
    g = obograph_util.convert_json_object(og)
    ont = Ontology(handle=None, payload=g)
    return ont

def translate_file_to_ontology(handle, **args):
    if handle.endswith(".json"):
        g = obograph_util.convert_json_file(handle, **args)
        return Ontology(handle=handle, payload=g)
    elif handle.endswith(".ttl"):
        from ontobio.sparql.rdf2nx import RdfMapper
        logger.info("RdfMapper: {}".format(args))
        m = RdfMapper(**args)
        return m.convert(handle,'ttl')
    else:
        if not (handle.endswith(".obo") or handle.endswith(".owl")):
            logger.info("Attempting to parse non obo or owl file with owltools: "+handle)
        encoded = get_checksum(handle)
        logger.info(" encoded: "+str(encoded))
        fn = '/tmp/'+encoded
        if not os.path.isfile(fn):
            cmd = ['owltools',handle,'-o','-f','json',fn]
            cp = subprocess.run(cmd, check=True)
            logger.info(cp)
        else:
            logger.info("using cached file: "+fn)
        g = obograph_util.convert_json_file(fn, **args)
        return Ontology(handle=handle, payload=g)

def get_checksum(file):
    """
    Get SHA256 hash from the contents of a given file
    """
    with open(file, 'rb') as FH:
        contents = FH.read()
    return hashlib.sha256(contents).hexdigest()



#!/usr/bin/env python3

import click
import os

from ontobio.rdfgen import assoc_rdfgen
from ontobio.io import assocparser
from ontobio.io import gafparser
from ontobio import ontol_factory

@click.group()
def cli():
    pass

@cli.command()
@click.option("--association", "-a", default="gaf", type=click.Choice(["gaf"]))
@click.option("--ontology", "-r", type=click.Path(exists=True, readable=True, dir_okay=False))
@click.option("--output", "-o", type=click.File('wb'))
@click.argument("association_file", type=click.Path(exists=True))
def convert(association, ontology, output, association_file):
    click.echo("converting {}".format(association))

    rdfWriter = assoc_rdfgen.TurtleRdfWriter(label=os.path.basename(output.name))
    rdfTransformer = assoc_rdfgen.CamRdfTransform(writer=rdfWriter)
    parser_config = assocparser.AssocParserConfig(ontology=make_ontology(ontology))
    parser = _association_parser(association, parser_config)

    with open(association_file) as af:
        lines = sum(1 for line in af)

    with open(association_file) as af:
        associations = parser.association_generator(file=af)
        with click.progressbar(iterable=associations, length=lines) as assocs:
            for assoc in assocs:
                rdfTransformer.provenance()
                rdfTransformer.translate(assoc)

        click.echo("Writing ttl to disk")
        rdfWriter.serialize(destination=output)


def _association_parser(association_type, config):
    if association_type == "gaf":
        return gafparser.GafParser(config=config)

def make_ontology(input_ontology):
    click.echo("Loading ontology...")
    ontology_factory = ontol_factory.OntologyFactory()
    return ontology_factory.create(input_ontology)


if __name__ == "__main__":
    cli()



from ontobio.assocmodel import AssociationSet


def jaccard_similarity(aset:AssociationSet, s1:str, s2:str) -> float:
    """
    Calculate jaccard index of inferred associations of two subjects

    |ancs(s1) /\ ancs(s2)|
    ---
    |ancs(s1) \/ ancs(s2)|

    """
    a1 = aset.inferred_types(s1)
    a2 = aset.inferred_types(s2)
    num_union = len(a1.union(a2))
    if num_union == 0:
        return 0.0
    return len(a1.intersection(a2)) / num_union
